@@ -1,42 +1,69 @@
 #include <string.h>
 #include "utils.h"
 
-#if ! AT_LEAST_LIBGIT2(0, 21)
-#define git_checkout_options git_checkout_opts
-#define GIT_CHECKOUT_OPTIONS_INIT GIT_CHECKOUT_OPTS_INIT
-#endif
-
 SEXP R_git_reset(SEXP ptr, SEXP ref, SEXP typenum){
-  git_object *revision = NULL;
   git_repository *repo = get_git_repository(ptr);
-  bail_if(git_revparse_single(&revision, repo, CHAR(STRING_ELT(ref, 0))), "git_revparse_single");
+  git_object *revision = resolve_refish(ref, repo);
   git_checkout_options opts = GIT_CHECKOUT_OPTIONS_INIT;
   opts.checkout_strategy = GIT_CHECKOUT_SAFE;
+  set_checkout_notify_cb(&opts);
   git_reset_t reset_type = Rf_asInteger(typenum);
   bail_if(git_reset(repo, revision, reset_type, &opts), "git_reset");
   return ptr;
 }
 
+static int branch_exists(git_repository *repo, const char *name, git_branch_t branch_type){
+  git_reference *ref = NULL;
+  if(git_branch_lookup(&ref, repo, name, branch_type) == GIT_OK){
+    git_reference_free(ref);
+    return 1;
+  }
+  return 0;
+}
+
+SEXP R_git_branch_current(SEXP ptr){
+  git_repository *repo = get_git_repository(ptr);
+  git_reference *head = NULL;
+  if(git_repository_head(&head, repo) == GIT_OK){
+    SEXP out = safe_string(git_reference_shorthand(head));
+    git_reference_free(head);
+    return out;
+  }
+  return R_NilValue;
+}
+
+SEXP R_git_branch_exists(SEXP ptr, SEXP name, SEXP local){
+  git_repository *repo = get_git_repository(ptr);
+  git_branch_t type = Rf_asLogical(local) ? GIT_BRANCH_LOCAL : GIT_BRANCH_REMOTE;
+  return Rf_ScalarLogical(branch_exists(repo, CHAR(STRING_ELT(name, 0)), type));
+}
+
 SEXP R_git_create_branch(SEXP ptr, SEXP name, SEXP ref, SEXP checkout){
   git_object *obj;
   git_commit *commit = NULL;
-  git_object *revision = NULL;
   git_reference *branch = NULL;
+  const char *source = CHAR(STRING_ELT(ref, 0));
   git_checkout_options opts = GIT_CHECKOUT_OPTIONS_INIT;
   opts.checkout_strategy = GIT_CHECKOUT_SAFE;
+  set_checkout_notify_cb(&opts);
   git_repository *repo = get_git_repository(ptr);
-  bail_if(git_revparse_single(&revision, repo, CHAR(STRING_ELT(ref, 0))), "git_revparse_single");
+  git_object *revision = resolve_refish(ref, repo);
   bail_if(git_commit_lookup(&commit, repo, git_object_id(revision)), "git_commit_lookup");
   git_object_free(revision);
   bail_if(git_branch_create(&branch, repo, CHAR(STRING_ELT(name, 0)), commit, 0), "git_branch_create");
   git_commit_free(commit);
+  if(branch_exists(repo, source, GIT_BRANCH_REMOTE)){
+    bail_if(git_branch_set_upstream(branch, source), "git_branch_set_upstream");
+  }
   if(Rf_asInteger(checkout)){
     bail_if(git_object_lookup(&obj, repo, git_reference_target(branch), GIT_OBJ_ANY), "git_object_lookup");
     bail_if(git_checkout_tree(repo, obj, &opts), "git_checkout_tree");
     git_object_free(obj);
     bail_if(git_repository_set_head(repo, git_reference_name(branch)), "git_repository_set_head");
   }
-  return ptr;
+  SEXP out = safe_string(git_reference_name(branch));
+  git_reference_free(branch);
+  return out;
 }
 
 SEXP R_git_delete_branch(SEXP ptr, SEXP branch){
@@ -56,6 +83,7 @@ SEXP R_git_checkout_branch(SEXP ptr, SEXP branch, SEXP force){
   /* Set checkout options */
   git_checkout_options opts = GIT_CHECKOUT_OPTIONS_INIT;
   opts.checkout_strategy = Rf_asLogical(force) ? GIT_CHECKOUT_FORCE : GIT_CHECKOUT_SAFE;
+  set_checkout_notify_cb(&opts);
 
   git_object *obj;
   bail_if(git_object_lookup(&obj, repo, git_reference_target(ref), GIT_OBJ_ANY), "git_object_lookup");
@@ -69,15 +97,14 @@ SEXP R_git_checkout_ref(SEXP ptr, SEXP ref, SEXP force){
   git_repository *repo = get_git_repository(ptr);
   git_checkout_options opts = GIT_CHECKOUT_OPTIONS_INIT;
   opts.checkout_strategy = Rf_asLogical(force) ? GIT_CHECKOUT_FORCE : GIT_CHECKOUT_SAFE;
+  set_checkout_notify_cb(&opts);
 
   /* Parse the branch/tag/ref string */
-  git_object *treeish = NULL;
-  const char *refstring = CHAR(STRING_ELT(ref, 0));
-  bail_if(git_revparse_single(&treeish, repo, refstring), "git_revparse_single");
+  git_object *treeish = resolve_refish(ref, repo);
   bail_if(git_checkout_tree(repo, treeish, &opts), "git_checkout_tree");
   git_object_free(treeish);
   char buf[1000];
-  snprintf(buf, 999, "refs/heads/%s", refstring);
+  snprintf(buf, 999, "refs/heads/%s", CHAR(STRING_ELT(ref, 0)));
   bail_if(git_repository_set_head(repo, buf), "git_repository_set_head");
   return ptr;
 }
@@ -120,48 +147,66 @@ SEXP R_git_branch_list(SEXP ptr){
   return build_tibble(5, "name", names, "local", islocal, "ref", refs,"upstream", upstreams, "commit", ids);
 }
 
-static SEXP make_refspecs(git_remote *remote){
-  int size = git_remote_refspec_count(remote);
-  SEXP out = PROTECT(Rf_allocVector(STRSXP, size));
-  for(int i = 0; i < size; i++){
-    SET_STRING_ELT(out, i, safe_char(git_refspec_string(git_remote_get_refspec(remote, i))));
-  }
-  UNPROTECT(1);
-  return out;
-}
-
 SEXP R_git_remote_list(SEXP ptr){
   git_strarray remotes = {0};
   git_repository *repo = get_git_repository(ptr);
   bail_if(git_remote_list(&remotes, repo), "git_remote_list");
   SEXP names = PROTECT(Rf_allocVector(STRSXP, remotes.count));
   SEXP url = PROTECT(Rf_allocVector(STRSXP, remotes.count));
-  SEXP refspecs = PROTECT(Rf_allocVector(VECSXP, remotes.count));
   for(int i = 0; i < remotes.count; i++){
     git_remote *remote = NULL;
     char *name = remotes.strings[i];
     SET_STRING_ELT(names, i, safe_char(name));
     if(!git_remote_lookup(&remote, repo, name)){
       SET_STRING_ELT(url, i, safe_char(git_remote_url(remote)));
-      SET_VECTOR_ELT(refspecs, i, make_refspecs(remote));
       git_remote_free(remote);
     }
     free(name);
   }
-  return build_tibble(3, "name", names, "url", url, "refspecs", refspecs);
+  return build_tibble(2, "name", names, "url", url);
 }
 
 SEXP R_git_remote_add(SEXP ptr, SEXP name, SEXP url, SEXP refspec){
   const char *curl = CHAR(STRING_ELT(url, 0));
   const char *cname = CHAR(STRING_ELT(name, 0));
-  const char *crefspec = Rf_length(refspec) ? CHAR(STRING_ELT(refspec, 0)) : NULL;
   git_repository *repo = get_git_repository(ptr);
   if(!git_remote_is_valid_name(cname))
     Rf_error("Invalid remote name %s", cname);
   git_remote *remote = NULL;
-  bail_if(git_remote_create_with_fetchspec(&remote,repo, cname, curl, crefspec), "git_remote_create");
+  if(Rf_length(refspec)){
+    const char *crefspec = CHAR(STRING_ELT(refspec, 0));
+    bail_if(git_remote_create_with_fetchspec(&remote, repo, cname, curl, crefspec), "git_remote_create_with_fetchspec");
+  } else {
+    bail_if(git_remote_create(&remote, repo, cname, curl), "git_remote_create");
+  }
+  SEXP out = safe_string(git_remote_name(remote));
   git_remote_free(remote);
-  return make_refspecs(remote);
+  return out;
+}
+
+SEXP R_git_remote_set_url(SEXP ptr, SEXP name, SEXP url){
+  git_remote * remote = NULL;
+  const char *curl = Rf_length(url) ? CHAR(STRING_ELT(url, 0)) : NULL;
+  const char *cname = CHAR(STRING_ELT(name, 0));
+  git_repository *repo = get_git_repository(ptr);
+  /* git_remote_lookup validates that the remote exists */
+  bail_if(git_remote_lookup(&remote, repo, cname), "git_remote_lookup");
+  bail_if(git_remote_set_url(repo, cname, curl), "git_remote_set_url");
+  SEXP out = safe_string(git_remote_url(remote));
+  git_remote_free(remote);
+  return out;
+}
+
+SEXP R_git_remote_set_pushurl(SEXP ptr, SEXP name, SEXP url){
+  git_remote * remote = NULL;
+  const char *curl = Rf_length(url) ? CHAR(STRING_ELT(url, 0)) : NULL;
+  const char *cname = CHAR(STRING_ELT(name, 0));
+  git_repository *repo = get_git_repository(ptr);
+  bail_if(git_remote_lookup(&remote, repo, cname), "git_remote_lookup");
+  bail_if(git_remote_set_pushurl(repo, cname, curl), "git_remote_set_url");
+  SEXP out = safe_string(git_remote_pushurl(remote));
+  git_remote_free(remote);
+  return out;
 }
 
 SEXP R_git_remote_remove(SEXP ptr, SEXP name){
@@ -171,7 +216,7 @@ SEXP R_git_remote_remove(SEXP ptr, SEXP name){
   return R_NilValue;
 }
 
-SEXP R_git_branch_set_upsteam(SEXP ptr, SEXP remote, SEXP branch){
+SEXP R_git_branch_set_upstream(SEXP ptr, SEXP upstream, SEXP branch){
   git_reference *ref;
   git_repository *repo = get_git_repository(ptr);
   if(Rf_length(branch)){
@@ -179,7 +224,73 @@ SEXP R_git_branch_set_upsteam(SEXP ptr, SEXP remote, SEXP branch){
   } else {
     bail_if(git_repository_head(&ref, repo), "git_repository_head");
   }
-  bail_if(git_branch_set_upstream(ref, CHAR(STRING_ELT(remote, 0))), "git_branch_set_upstream");
+  bail_if(git_branch_set_upstream(ref, CHAR(STRING_ELT(upstream, 0))), "git_branch_set_upstream");
   git_reference_free(ref);
   return ptr;
 }
+
+SEXP R_git_remote_refspecs(SEXP ptr, SEXP name){
+  git_remote *remote = NULL;
+  const char *cname = CHAR(STRING_ELT(name, 0));
+  git_repository *repo = get_git_repository(ptr);
+  bail_if(git_remote_lookup(&remote, repo, cname), "git_remote_lookup");
+  size_t len = git_remote_refspec_count(remote);
+  SEXP names = PROTECT(Rf_allocVector(STRSXP, len));
+  SEXP urls = PROTECT(Rf_allocVector(STRSXP, len));
+  SEXP directions = PROTECT(Rf_allocVector(STRSXP, len));
+  SEXP string = PROTECT(Rf_allocVector(STRSXP, len));
+  SEXP src = PROTECT(Rf_allocVector(STRSXP, len));
+  SEXP dest = PROTECT(Rf_allocVector(STRSXP, len));
+  SEXP force = PROTECT(Rf_allocVector(LGLSXP, len));
+  for(size_t i = 0; i < len; i++){
+    const git_refspec *refspec = git_remote_get_refspec(remote, i);
+    SET_STRING_ELT(names, i, safe_char(git_remote_name(remote)));
+    SET_STRING_ELT(urls, i, safe_char(git_remote_url(remote)));
+    SET_STRING_ELT(directions, i, safe_char(git_refspec_direction(refspec) == GIT_DIRECTION_FETCH ? "fetch" : "push"));
+    SET_STRING_ELT(string, i, safe_char(git_refspec_string(refspec)));
+    SET_STRING_ELT(src, i, safe_char(git_refspec_src(refspec)));
+    SET_STRING_ELT(dest, i, safe_char(git_refspec_dst(refspec)));
+    LOGICAL(force)[i] = git_refspec_force(refspec);
+  }
+  return build_tibble(7, "name", names, "url", urls, "direction", directions,
+                      "refspec", string, "src", src, "dest", dest, "force", force);
+}
+
+SEXP R_git_remote_add_fetch(SEXP ptr, SEXP remote, SEXP refspec){
+  git_repository *repo = get_git_repository(ptr);
+  bail_if(git_remote_add_fetch(repo, CHAR(STRING_ELT(remote, 0)), CHAR(STRING_ELT(refspec, 0))), "git_remote_add_fetch");
+  return refspec;
+}
+
+SEXP R_git_remote_info(SEXP ptr, SEXP name){
+  git_remote *remote = NULL;
+  const char *cname = CHAR(STRING_ELT(name, 0));
+  git_repository *repo = get_git_repository(ptr);
+  bail_if(git_remote_lookup(&remote, repo, cname), "git_remote_lookup");
+  git_strarray fetchspecs = {0};
+  git_strarray pushspecs = {0};
+  bail_if(git_remote_get_fetch_refspecs(&fetchspecs, remote), "git_remote_get_fetch_refspecs");
+  bail_if(git_remote_get_push_refspecs(&pushspecs, remote), "git_remote_get_push_refspecs");
+  SEXP fetch = PROTECT(Rf_allocVector(STRSXP, fetchspecs.count));
+  SEXP push = PROTECT(Rf_allocVector(STRSXP, pushspecs.count));
+  for(int i = 0; i < fetchspecs.count; i++)
+    SET_STRING_ELT(fetch, i, safe_char(fetchspecs.strings[i]));
+  for(int i = 0; i < pushspecs.count; i++)
+    SET_STRING_ELT(push, i, safe_char(pushspecs.strings[i]));
+  git_strarray_free(&fetchspecs);
+  git_strarray_free(&pushspecs);
+  char buf[1000] = {0};
+  sprintf(buf, "refs/remotes/%s/HEAD", git_remote_name(remote));
+  git_reference *remote_head = NULL;
+  int has_default = git_reference_lookup(&remote_head, repo, buf) == GIT_OK;
+  SEXP out = build_list(6,
+    "name", PROTECT(string_or_null(git_remote_name(remote))),
+    "url", PROTECT(string_or_null(git_remote_url(remote))),
+    "push_url", PROTECT(string_or_null(git_remote_pushurl(remote))),
+    "head", PROTECT(string_or_null(has_default ? git_reference_symbolic_target(remote_head) : NULL)),
+    "fetch", fetch,
+    "push", push
+  );
+  git_remote_free(remote);
+  return out;
+}
